﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using KludgeBox.Collections;

namespace VectorSurvivors;


public readonly record struct Tier(string Name, int Value, Color Color)
{
    public static readonly Tier Common = new("Обычное", 1, Colors.White);
    public static readonly Tier Uncommon = new("Необычное", 2, Colors.Green.Lightened(0.2f));
    public static readonly Tier Rare = new("Редкое", 3, Colors.Cyan.Lightened(0.2f));
    public static readonly Tier Epic = new("Эпическое", 4, Colors.Purple.Lightened(0.4f));
    public static readonly Tier Legendary = new("Легендарное", 5, Colors.Red.Lightened(0.2f));
    
    /// <summary>
    /// Can't be generated by upgrades GetRandomTier() directly, but may be replaced after passing into actual IUpgradable
    /// </summary>
    public static readonly Tier Unique = new("Уникальное", 0, Colors.Gold);
    
    /// <summary>
    /// Special tier for obtaining new weapons
    /// </summary>
    public static readonly Tier Weapon = new("Новое оружие", 0, Colors.OrangeRed);
    
    
    /// <inheritdoc />
    public override string ToString()
    {
        return Name;
    }
    
    public static implicit operator int(Tier tier) => tier.Value;
}

public delegate PendingUpgrade ApplyUpgradeDelegate(Player player,Tier tier);

public static class UpgradeUtils
{
    public static ReadOnlyDictionary<Tier, double> TierWeights => _tierWeights.AsReadOnly();
    public static ReadOnlyHashSet<Tier> DefaultTiers => _tiers.AsReadOnly();


    private static readonly HashSet<Tier> _tiers = new()
    {
        Tier.Common,
        Tier.Uncommon,
        Tier.Rare,
        Tier.Epic,
        Tier.Legendary
    };
    
    private static readonly Dictionary<Tier, double> _tierWeights = new()
    {
        {Tier.Common, 50},
        {Tier.Uncommon, 25},
        {Tier.Rare, 12},
        {Tier.Epic, 8},
        {Tier.Legendary, 5}
    };

    private static GuaranteedWeightedRandom<Tier> _tierRandom = new();
    
    
    static UpgradeUtils()
    {
        foreach ((Tier tier, var weight) in _tierWeights)
        {
            _tierRandom.Add(tier, weight);
        }
    }
    
    public static Tier GetRandomTier()
    {
        return _tierRandom.PickRandom();
    }
    
    public static double SwitchTier(Tier tier, double common, double uncommon, double rare, double epic, double legendary)
    {
        return (int)tier switch
        {
            1 => common,
            2 => uncommon,
            3 => rare,
            4 => epic,
            5 => legendary,
            _ => throw new ArgumentOutOfRangeException(nameof(tier), tier, null)
        };
    }
}