﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;

namespace VectorSurvivors;


/// <summary>
/// Collection of upgrade delegates that can store ApplyUpgradeDelegate for specific context.
/// Recommended to use within specific IUpgradable.
/// </summary>
public class UpgradesRepository
{
    private Dictionary<Tier, List<ApplyUpgradeDelegate>> _tiers = new();
    
    public UpgradesRepository()
    {
        foreach (Tier tier in UpgradeUtils.DefaultTiers)
        {
            _tiers.Add(tier, new List<ApplyUpgradeDelegate>());
        }
    }

    public ReadOnlyCollection<ApplyUpgradeDelegate> GetAllForTier(Tier tier)
    {
        return _tiers[tier].AsReadOnly();
    }
    
    public ApplyUpgradeDelegate GetAnyForTier(Tier tier)
    {
        return _tiers[tier].GetRandomOrDefault();
    }

    public void Add(ApplyUpgradeDelegate applyUpgrade, Tier tier)
    {
        _tiers[tier].Add(applyUpgrade);
    }
    
    public void Add(ApplyUpgradeDelegate applyUpgrade, params Tier[] tiers)
    {
        foreach (Tier tier in tiers)
        {
            Add(applyUpgrade, tier);
        }
    }
    
    public void AddToAll(ApplyUpgradeDelegate applyUpgrade)
    {
        foreach (Tier tier in UpgradeUtils.DefaultTiers)
        {
            Add(applyUpgrade, tier);
        }
    }
    
    public void AddToAllExcept(ApplyUpgradeDelegate applyUpgrade, Tier tier)
    {
        foreach (Tier t in UpgradeUtils.DefaultTiers)
        {
            if (t == tier) continue;
            Add(applyUpgrade, t);
        }
    }

    public void AddToAllExcept(ApplyUpgradeDelegate applyUpgrade, params Tier[] tiers)
    {
        foreach (Tier tier in UpgradeUtils.DefaultTiers)
        {
            Add(applyUpgrade, tier);
        }
    }

    public void Remove(ApplyUpgradeDelegate applyUpgrade)
    {
        foreach (Tier tier in UpgradeUtils.DefaultTiers)
        {
            _tiers[tier].Remove(applyUpgrade);
        }
    }
}