using System;
using System.Linq;
using Godot.Collections;
using KludgeBox.Scheduling;
using VectorSurvivors.Events;

namespace VectorSurvivors;

public partial class Character : CharacterBody2D
{
	[ExportGroup("Configuration")]
	[Export] [NotNull] public Sprite2D Sprite { get; private set; }
	[Export] [NotNull] public Smoothing2D Smoother { get; private set; }
	[Export] public CharacterData Data { get; private set; }

	[ExportGroup("State")] 
	[Export] public double Health { get; set; } = -1;
	
	[ExportGroup("Logic")]
	[Export] public CharacterController Controller { get; internal set; }
	[Export] public Array<CharacterComponent> PredefinedComponents { get; private set; } = new();
	
	public double HealthFraction => Health / MaxHealth;

	
	// Stats
	public double MaxHealth => AdjustValue((CharacterComponent c, ref double v) => c.AdjustMaxHealth(this, ref v), Data.MaxHealth);
	public double HealthRegen => AdjustValue((CharacterComponent c, ref double v) => c.AdjustHealthRegen(this, ref v), Data.HealthRegen);
	public int XpReward => AdjustIntValue((CharacterComponent c, ref int v) => c.AdjustXpReward(this, ref v), Data.XpReward);
	public int MoneyReward => AdjustIntValue((CharacterComponent c, ref int v) => c.AdjustMoneyReward(this, ref v), Data.MoneyReward);
	
	// Movement
	public double AccelerationTime => AdjustValue((CharacterComponent c, ref double v) => c.AdjustAccelerationTime(this, ref v), Data.AccelerationTime);
	public double MovementSpeed => AdjustValue((CharacterComponent c, ref double v) => c.AdjustMovementSpeed(this, ref v), Data.MovementSpeed);
	public double Friction => AdjustValue((CharacterComponent c, ref double v) => c.AdjustFriction(this, ref v), Data.Friction);
	public double RotationSpeed => AdjustValue((CharacterComponent c, ref double v) => c.AdjustRotationSpeed(this, ref v), Data.RotationSpeed);
	
	// Attack
	public double UniversalAttackDamageFactor => AdjustValue((CharacterComponent c, ref double v) => c.AdjustUniversalAttackDamageFactor(this, ref v), Data.UniversalAttackDamageFactor);
	public double UniversalAttackSpeedFactor => AdjustValue((CharacterComponent c, ref double v) => c.AdjustUniversalAttackSpeedFactor(this, ref v), Data.UniversalAttackSpeedFactor);
	public double UniversalAttackRangeFactor => AdjustValue((CharacterComponent c, ref double v) => c.AdjustUniversalAttackRangeFactor(this, ref v), Data.UniversalAttackRangeFactor);

	public bool IsImmobilized => Mathf.Abs(MovementSpeed) < 0.01;
	protected delegate void AdjustDoubleDelegate(CharacterComponent component, ref double value);
	protected delegate void AdjustIntDelegate(CharacterComponent component, ref int value);

	protected double _hitFlash = 0;
	protected double _hitFlashDuration = 0.2;
	protected double AdjustValue(AdjustDoubleDelegate adjustDelegate, double value)
	{
		var tmpValue = value;
		foreach ((Type _, CharacterComponent component) in Components)
		{
			adjustDelegate(component, ref tmpValue);
		}
		
		return tmpValue;
	}
	protected int AdjustIntValue(AdjustIntDelegate adjustDelegate, int value)
	{
		var tmpValue = value;
		foreach ((Type _, CharacterComponent component) in Components)
		{
			adjustDelegate(component, ref tmpValue);
		}
		
		return tmpValue;
	}

	private Cooldown _regenCd = new(1);
	


	protected Character()
	{
		TreeExiting += CleanUp;
		_regenCd.Ready += () =>
		{
			Health += HealthRegen;
			if (Health > MaxHealth)
			{
				Health = MaxHealth;
			}
		};
	}
	
	/// <inheritdoc />
	public override void _Ready()
	{
		NotNullChecker.CheckProperties(this);
		ApplyPredefinedComponents();
		ApplyPreset(Data);
		EventBus.Publish(new CharacterSpawnedEvent(this));
	}

	/// <inheritdoc />
	public override void _Process(double delta)
	{
		_regenCd.Update(delta);
		Controller?.OnProcess(this, delta);
		foreach ((Type _, CharacterComponent value) in Components)
		{
			value.OnProcess(this, delta);
		}
		
		Sprite.SetShaderParameter("colorMaskFactor", _hitFlash);
		_hitFlash = Mathf.Max(0, _hitFlash - delta / _hitFlashDuration);
		RotateToAimPos(delta);
	}

	/// <inheritdoc />
	public override void _PhysicsProcess(double delta)
	{
		var inputVector = Controller?.GetInputVector(this) ?? Vec();
		
		ProcessMovement(inputVector.Normalized(), delta);
	}


	public void ApplyPreset(CharacterData data)
	{
		Data = data;
		Health = Mathf.Min(Health, data.MaxHealth);
		Health = Health < 0 ? data.MaxHealth : Health;
		if (!String.IsNullOrWhiteSpace(data.SpriteOverridePath))
		{
			var texture = GD.Load<Texture2D>(data.SpriteOverridePath);
			if (texture.IsValid())
			{
				Sprite.Texture = texture;
			}
			else
			{
				Log.Error($"Failed to load sprite override '{data.SpriteOverridePath}'");
			}
		}
	}

	public virtual void CleanUp()
	{
		var comps = Components.ToArray();
		foreach ((Type key, CharacterComponent value) in comps)
		{
			RemoveComponent(value);
		}
	}

	public virtual bool TakeDamage(Damage damage)
	{
		if(EventBus.TryPublish(new CharacterTakingDamageEvent(this, damage))) return false;

		foreach ((Type _, CharacterComponent component) in Components)
		{
			component.TakingDamage(this, damage);
		}

		if(damage.IsIgnored && damage.CanBeIgnored) return false;
		_hitFlash = 1;
		
		Health -= damage.Amount;
		var label = FloatingLabel.Create();
		var scale = 0.8;//Mathf.Min(Math.Log(damage.Amount, 10), 0.5f);
		label.Configure($"{Mathf.FloorToInt(damage.Amount)}", damage.IsCritical ? Colors.Red : Colors.LightGray, scale);
		label.Position = Position + Rand.InsideUnitCircle * 20;
		
		
		World.AddEntity(label);
		if (Health <= 0)
		{
			EventBus.Publish(new CharacterDiedEvent(this, damage));
			World.Notify_CharacterDied(this);
		}

		return true;
	}
	
	
	private void ProcessMovement(Vector2 inputVector, double delta)
	{
		var friction = Friction * World.Instance.Friction;
		
		var speed = Velocity.Length();
		var acceleration = MovementSpeed / AccelerationTime * friction;
		var breaking = Mathf.Min(speed, acceleration * delta);

		speed += acceleration * delta * 2;

		var velDir = Velocity.Normalized();
		Velocity += inputVector * acceleration * delta * 2 - velDir * breaking;

		Velocity = Velocity.LimitLength(MovementSpeed);

		MoveAndSlide();
	}
	
	public void RotateToAimPos(double delta)
	{
		//Куда хотим повернуться
		double targetAngle = GetAngleToAimPos();
		//На какой угол надо повернуться (знак указывает направление)
		double deltaAngleToTargetAngel = Mathf.AngleDifference(Rotation - Mathf.Pi / 2, targetAngle);
		//Только направление (-1, 0, 1)
		double directionToTargetAngel = Mathf.Sign(deltaAngleToTargetAngel);
		//Максимальная скорость поворота (за секунду)
		double rotationSpeedRad = RotationSpeed;
		//Максимальная скорость поворота (за прошедшее время)
		rotationSpeedRad *= delta;
		//Если надо повернуться на угол меньший максимальной скорости, то обрезаем скорость, чтобы повернуться ровно в цель
		rotationSpeedRad = Mathf.Min(rotationSpeedRad, Mathf.Abs(deltaAngleToTargetAngel));
		//Добавляем к скорости поворота направление, чтобы поворачивать в сторону цели
		rotationSpeedRad *= directionToTargetAngel;
		//Поворачиваемся на угол
		Rotation += rotationSpeedRad;
	}
	
	public double GetAngleToAimPos()
	{
		// Получаем текущую позицию мыши
		var mousePos = Controller?.GetAimPos(this) ?? GlobalPosition;
		// Вычисляем вектор направления от объекта к мыши
		var mouseDir = GlobalPosition.DirectionTo(mousePos);
		// Вычисляем направление от объекта к мыши
		return mouseDir.Angle();
	}
	
	[Rpc(MultiplayerApi.RpcMode.Authority, TransferMode = MultiplayerPeer.TransferModeEnum.Reliable, CallLocal = true)]
	private void SyncPreset(string presetJson)
	{
		var preset = CharacterData.FromJson(presetJson);
		ApplyPreset(preset);
	}
}