﻿using System;

namespace VectorSurvivors;

[GlobalClass]
public partial class UpgradeComponent : CharacterComponent, IUpgradable
{
    /// <inheritdoc />

    public string UpgradableName => "Игрок";
    public string Description => "Улучшения персонажа игрока";

    public string StateInfo
    {
        get
        {
            return $"Макс. здоровье: {_player.MaxHealth:N0}\n" +
                   $"Скорость передвижения: {_player.MovementSpeed:N0}\n" +
                   $"Радиус сбора сфер: {_player.OrbCollectionRange:N0}\n" +
                   $"Регенерация здоровья: {_player.HealthRegen:N1}";
        }
    }
    public int Level { get; set; } = 0;
    public int MaxLevel => World.Instance.MaxPlayerLevel;
    
    public double MaxHpOffset { get; private set; } = 0.0;
    public double MovementSpeedOffset { get; private set; } = 0.0;
    public double OrbCollectionRangeOffset { get; private set; } = 0.0;
    public double HealthRegenOffset { get; private set; } = 0.0;
    
    private UpgradesRepository _upgrades = new();
    public Texture2D Icon => Game.Storage.Textures.CharacterIcon.Duplicate() as Texture2D;
    private Player _player;

    public UpgradeComponent()
    {
        _upgrades.AddToAll(MaxHpUpgrade);
        _upgrades.AddToAll(MovementSpeedUpgrade);
        _upgrades.AddToAll(OrbCollectionRangeUpgrade);
        _upgrades.AddToAll(HealthRegenUpgrade);
        
        _upgrades.Add(GetWeaponUpgrade, Tier.Legendary, Tier.Epic);
    }

    /// <inheritdoc />
    public override void OnAdded(Character character)
    {
        if(character is Player player)
            _player = player;
    }

    private PendingUpgrade GetWeaponUpgrade(Player player, Tier tier)
    {
        if (!player.Weapons.CanAddWeapon())
        {
            return TryGetAnotherUpgrade(player, tier);
        }
        
        var newWeaponType = player.Weapons.GetMissingWeaponTypes().GetRandom();
        tier = Tier.Weapon;
        
        var tmpNewWeapon = Game.Storage.Weapons.ByType[newWeaponType].Instantiate() as Weapon;
        var name = tmpNewWeapon.UpgradableName;
        var description = tmpNewWeapon.Description;
        var icon = tmpNewWeapon.Icon;
        
        tmpNewWeapon.QueueFree();
        var addWeaponAction = () =>
        {
            var newWeapon = Game.Storage.Weapons.ByType[newWeaponType].Instantiate() as Weapon;
            player.Weapons.AddWeapon(newWeapon);
        };

        return new PendingUpgrade(name, description, tier, this, addWeaponAction, icon);
    }


    /// <inheritdoc />
    public virtual PendingUpgrade GetUpgrade(Player player, Tier tier)
    {
        return _upgrades.GetAnyForTier(tier)?.Invoke(player, tier);
    }

    /// <inheritdoc />
    public bool CanBeUpgraded()
    {
        return Level < World.Instance.MaxPlayerLevel;
    }

    /// <inheritdoc />
    public void OnUpgrade(PendingUpgrade upgrade)
    {
        Level++;
    }

    /// <inheritdoc />
    public override void AdjustOrbCollectionRange(Character character, ref double value)
    {
        value += OrbCollectionRangeOffset;
    }
    
    /// <inheritdoc />
    public override void AdjustMaxHealth(Character character, ref double value)
    {
        value += MaxHpOffset;
    }
    
    /// <inheritdoc />
    public override void AdjustMovementSpeed(Character character, ref double speed)
    {
        speed += MovementSpeedOffset;
    }
    
    public override void AdjustHealthRegen(Character character, ref double value)
    {
        value += HealthRegenOffset;
    }

    private PendingUpgrade MovementSpeedUpgrade(Player player, Tier tier)
    {
        var baseSpeed = 25;
        var tieredSpeed = baseSpeed * tier;
        var oldSpeed = _player.MovementSpeed;
        var newSpeed = _player.MovementSpeed + tieredSpeed;

        var name = $"Скорость передвижения +{tieredSpeed}";
        var description =
            $"Увеличивает скорость передвижения\n {oldSpeed:N0} -> {newSpeed:N0} {TextUtils.Declension((int)newSpeed, "пиксель", "пикселя", "пикселей")} в секунду.";
        
        return new PendingUpgrade(name, description, tier, this, () => MovementSpeedOffset += tieredSpeed, Icon);
    }
    
    private PendingUpgrade OrbCollectionRangeUpgrade(Player player, Tier tier)
    {
        var baseRange = 15;
        var tieredRange = baseRange * tier;
        var oldRange = _player.OrbCollectionRange;
        var newRange = _player.OrbCollectionRange + tieredRange;
        
        var name = $"Радиус сбора сфер +{tieredRange:N0}";
        var description = $"Увеличивает радиус сбора сфер\n{oldRange:N0} -> {newRange:N0} {TextUtils.Declension((int)newRange, "пиксель", "пикселя", "пикселей")}.";
        
        return new PendingUpgrade(name, description, tier, this, () => OrbCollectionRangeOffset += tieredRange, Icon);
    }


    private PendingUpgrade MaxHpUpgrade(Player player, Tier tier)
    {
        var baseValue = 2;
        var value = UpgradeUtils.SwitchTier(tier,
            common: baseValue,
            uncommon: baseValue * 2,
            rare: baseValue * 3,
            epic: baseValue * 5,
            legendary: baseValue * 10);
        
        var oldHp = player.Health;
        var newHp = player.Health + value;
        
        var name = $"Макс. здоровье +{value:N0}";
        var description = $"Увеличивает максимальное здоровье\n{oldHp:N0} -> {newHp:N0}";

        return new PendingUpgrade(name, description, tier, this, () =>
        {
            MaxHpOffset += value;
            player.Health += value;
        }, Icon);
    }
    
    private PendingUpgrade HealthRegenUpgrade(Player player, Tier tier)
    {
        var baseRegen = 0.1;
        var tieredRegen = (int)tier switch
        {
            1 => baseRegen,
            2 => baseRegen * 2,
            3 => baseRegen * 4,
            4 => baseRegen * 6,
            5 => baseRegen * 10,
            _ => throw new ArgumentOutOfRangeException(nameof(tier), tier, null)
        };
        
        var oldRegen = player.HealthRegen;
        var newRegen = player.HealthRegen + tieredRegen;
        
        var name = $"Регенерация здоровья +{tieredRegen:N1} HP/с";
        var description = $"Увеличивает регенерацию здоровья\n{tieredRegen:N1} -> {newRegen:N1} в секунду.";
        
        return new PendingUpgrade(name, description, tier, this, () => HealthRegenOffset += tieredRegen, Icon);
    }
    
    private int _getUpgradeAttempts = 0;
    private PendingUpgrade TryGetAnotherUpgrade(Player player, Tier tier)
    {
        if (_getUpgradeAttempts > 5)
        {
            return null;
        }
		
        _getUpgradeAttempts++;
        return GetUpgrade(player, tier);
    }
}