using System;
using System.Collections.Generic;
using KludgeBox.Godot.Net;
using VectorSurvivors.Events;

namespace VectorSurvivors;

[GlobalClass]
public partial class Player : Character
{
	[Export] public int PlayerId { get; internal set; }
	[Export] public int Level { get; private set; }
	[Export] public int Xp { get; private set; }
	[Export] private double _orbCollectionRange { get; set; } = 150;
	[Export] public int Rerolls { get; set; } = 1;
	
	[Export] public SmoothCamera2D Camera { get; private set; }
	[Export] public WeaponsInventory Weapons { get; private set; }
	[Export] public UpgradeComponent UpgradesComponent { get; private set; } = new();
	
	public bool CanLevelUp => Level < World.Instance.MaxPlayerLevel;

	
	public double OrbCollectionRange => AdjustValue((CharacterComponent c, ref double v) => c.AdjustOrbCollectionRange(this, ref v), _orbCollectionRange);
	public int RequiredXp => Mathf.CeilToInt(World.Instance.BaseRequiredXp 
	                               * Mathf.Pow(World.Instance.RequiredXpLevelFactor, Level)
	                               + (World.Instance.RequiredXpLevelOffset * (Level)));
	
	public bool IsLocal => PlayerId == Network.MyPeerId;
	// Called when the node enters the scene tree for the first time.
	public override void _EnterTree()
	{
		base._EnterTree();
		PlayerId = Int32.Parse(Name);
		SetMultiplayerAuthority(PlayerId);
	}
	
	public override void _Ready()
	{
		base._Ready();
		
		Threading.DeferTask(() =>
		{
			if (IsLocal)
			{
				EventBus.Publish(new LocalPlayerSpawnedEvent(this));
				AddComponent(UpgradesComponent);
			}
			else
			{
				Controller = null;
			}
		});
	}

	// Called every frame. 'delta' is the elapsed time since the previous frame.
	public override void _Process(double delta)
	{
		base._Process(delta);
		
		Weapons.Update(this, delta * UniversalAttackSpeedFactor);
	}

	/// <inheritdoc />
	public override void _PhysicsProcess(double delta)
	{
		base._PhysicsProcess(delta);

		foreach (Orb orb in World.Orbs)
		{
			if ((orb.Position - Position).LengthSquared() < OrbCollectionRange * OrbCollectionRange)
				if (!orb.IsPicked)
					orb.Pick(this);
		}
	}
	
	public void OnCameraDropped(Node node)
	{
		EventBus.Publish(new CameraDroppedEvent(node as SmoothCamera2D, this));
	}

	public void AddXp(int amount)
	{
		if(!CanLevelUp) return;
		
		Xp += amount;
		if (Xp >= RequiredXp)
		{
			LevelUp();
		}
	}

	private bool _firstLevel = true;
	public void LevelUp(bool takeXp = true)
	{
		if(!CanLevelUp) return;
		
		if (takeXp)
		{
			Xp -= RequiredXp;
		}
		var provider = new UpgradesProvider()
			.With(UpgradesComponent)
			.WithAll(Weapons);
		
		ShowLevelUpUi(_firstLevel, provider);
		_firstLevel = false;
		
		Level++;
		UpgradesComponent.Level = Level;
	}

	public void SelectWeapon()
	{
		ShowLevelUpUi(false, new WeaponsProvider());
	}

	public void ShowLevelUpUi(bool buffed, IUpgradesProvider provider)
	{
		var levelUp = Game.Storage.Ui.LevelUpScreen.Instantiate() as LevelUpScreen;
		
		levelUp.Player = this;
		levelUp.Buffed = buffed;
		levelUp.Provider = provider;
		
		Game.ShowUi(levelUp);
	}
	
	public override void CleanUp()
	{
		if (Camera.IsValid())
		{
			Camera.QueueFree();
		}
		base.CleanUp();
	}

	public List<IUpgradable> GetUpgradables()
	{
		List<IUpgradable> upgradables = new();
		upgradables.Add(UpgradesComponent);
		upgradables.AddRange(Weapons.Weapons);
		return upgradables;
	}
}
