﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace VectorSurvivors;

public partial class UiContainer : Node
{
    public IUiScene CurrentState { get; private set; }
    public IUiScene RootState => StateStack.Last();
    
	public bool IsValidState => (CurrentState as Node).IsValid();
    public Stack<IUiScene> StateStack { get; private set; } = new();
    
    
    /// <inheritdoc />
    public override void _Ready()
    {
	    if (GetChildCount() > 0)
	    {
		    var child = GetChildren()[0] as IUiScene;
		    CurrentState = child;
	    }
		    
    }
    

    public void PushState(IUiScene newState, Action<Node> callback = null)
    {
    	StateStack.Push(ChangeTo(newState, callback, false));
    }

    public IUiScene PopState(Action<Node> callback = null, bool forced = false)
    {
	    var currentNode = CurrentState as Node;
	    if (CurrentState.PreventClose && currentNode.IsValid() && !currentNode.IsQueuedForDeletion() && !forced) 
		    return CurrentState;
	    
    	if(StateStack.Count > 0)
    		return ChangeTo(StateStack.Pop(), callback);

    	return ChangeTo(null, callback);
    }
    
    public void ClearStack()
    {
	    foreach (IUiScene uiScene in StateStack)
	    {
		    if (uiScene is Node node && node.IsValid())
		    {
			    node.QueueFree();
		    }
	    }

	    StateStack.Clear();
    }
    

    /// <summary>
    /// Заменяет текущее состояние на новое и делает обратный вызов с новым состоянием. Возвращает ссылку на предыдущее состояние.
    /// </summary>
    /// <remarks>
    /// Возвращаемое предыдущее состояние уже находится в очереди на удаление.
    /// </remarks>
    /// <returns>Предыдущее состояние</returns>
    public IUiScene ChangeTo(IUiScene newState, Action<Node> callback = null, bool freeOldState = true)
    {
	    if(newState == CurrentState) return CurrentState;
		    
    	var oldState = CurrentState;
	    var oldNode = oldState as Node;
    	CurrentState = newState;
    	if (oldNode.IsValid())
    	{
		    if(oldState.PausesGame)
		    {
			    Game.Unpause();
		    }

		    if (freeOldState)
		    {
			    oldNode?.QueueFree();
		    }
		    
		    if(oldNode.IsInsideTree())
		    {
			    RemoveChild(oldNode);
		    }
	    }
	    
	    var newNode = newState as Node;
	    if (newNode.IsValid())
	    {
		    AddChild(newNode, true);
		    newNode.TreeExited += () =>
		    {
			    if (newNode.IsQueuedForDeletion() && newNode == CurrentState)
				    PopState();
		    };
	    }
	    
	    if(newNode.IsValid() && newState.PausesGame)
	    	Game.Pause();
    	
    	callback?.Invoke(newNode);
    	return oldState;
    }
    
    /// <summary>
    /// Заменяет текущее состояние на новое и делает обратный вызов с новым состоянием. Возвращает ссылку на предыдущее состояние.
    /// </summary>
    /// <remarks>
    /// Возвращаемое предыдущее состояние уже находится в очереди на удаление.
    /// </remarks>
    /// <returns>Предыдущее состояние</returns>
    internal IUiScene ChangeTo(PackedScene newState, Action<Node> callback = null)
    {
	    if (newState is null) return ChangeTo(null, null, true);
	    var newUi = newState.Instantiate() as IUiScene;
	    if (newUi is null)
		    throw new ArgumentException($"{newState} is not IUiScene or null");
	    
    	return ChangeTo(newUi, callback);
    }
}

public interface IUiScene
{
    bool PausesGame { get; }
    bool PreventClose { get; }
}