namespace VectorSurvivors;

public partial class LevelUpScreen : CanvasLayer, IUiScene
{
	[Export] [NotNull] private HBoxContainer CardsContainer { get; set; }
	[Export] [NotNull] private Button DropButton { get; set; }
	[Export] [NotNull] private Button RerollButton { get; set; }
	[Export] [NotNull] private Label SelectUpgradeLabel { get; set; }
	
	public bool PausesGame => true;
	public bool PreventClose => true;
	
	public Player Player { get; internal set; }
	public bool Buffed { get; set; }
	public int UpgradeRerolls { get; set; } = 1;
	public IUpgradesProvider Provider { get; set; }

	private bool _wasRerolled = false;
	
	// Called when the node enters the scene tree for the first time.
	public override void _Ready()
	{
		NotNullChecker.CheckProperties(this);
		FillCards();
		UpdateButtons();
		RerollButton.Pressed += OnReroll;
		DropButton.Pressed += OnDrop;
		UpgradeRerolls = World.Instance.MaxRerollsOnLevelUp;
		SelectUpgradeLabel.Text = Provider.NameOverride ?? SelectUpgradeLabel.Text;
	}

	public void AddCard(PendingUpgrade upgrade)
	{
		var card = Game.Storage.Ui.UpgradeCard.Instantiate() as UpgradeCard;
		card.UpgradeButton.Pressed += OnSelected;
		card.Initialize(upgrade);
		CardsContainer.AddChild(card);
	}

	private void FillCards()
	{
		var upgrades = Provider.PickUpgrades(Player, World.Instance.UpgradesPerLevel, Buffed);
		if (upgrades.Count <= 0) return;
		foreach (var upgrade in upgrades)
		{
			AddCard(upgrade);
		}
	}

	private void UpdateButtons()
	{
		if (!Provider.CanBeDropped)
		{
			RerollButton.Visible = false;
			DropButton.Visible = false;
		}
		else
		{
			RerollButton.Text = $"Повторы: {Mathf.Min(Player.Rerolls, UpgradeRerolls)} ({Player.Rerolls})";
			if (Player.Rerolls <= 0 || UpgradeRerolls <= 0)
			{
				RerollButton.Disabled = true;
			}
			else
			{
				RerollButton.Disabled = false;
			}

			if (_wasRerolled || Player.Level == World.Instance.MaxPlayerLevel - 1)
			{
				DropButton.Disabled = true;
			}
		}
	}

	private void OnSelected()
	{
		QueueFree();
	}
	
	private void OnDrop()
	{
		Player.Rerolls += World.Instance.RerollsPerDrop;
		QueueFree();
	}
	
	private void OnReroll()
	{
		if (Player.Rerolls > 0 && UpgradeRerolls > 0)
		{
			_wasRerolled = true;
			Player.Rerolls--;
			UpgradeRerolls--;

			foreach (Node child in CardsContainer.GetChildren())
			{
				child.QueueFree();
			}

			FillCards();
			UpdateButtons();
		}
	}
}