using System.Collections.Generic;
using Content;
using KludgeBox.Collections;
using KludgeBox.Godot.Net;
using KludgeBox.Structs;
using VectorSurvivors.Events;

namespace VectorSurvivors;

public partial class World : Node2D
{
	[Export] [NotNull] public Floor Floor { get; private set; }
	[Export] private double _gameTime { get; set; }
	
	private Player _localPlayer { get; set; }
	
	private readonly HashSet<Character> _characters = new();
	private readonly HashSet<Enemy> _enemies = new();
	private readonly HashSet<Player> _players = new();
	private readonly HashSet<Orb> _orbs = new();
	private readonly HashSet<Node2D> _entities = new();

	private List<ListenerToken> _tokens = new();
	private WeightedRandom<PackedScene> _spawnPool = new();

	public override void _Ready()
	{
		NotNullChecker.CheckProperties(this);
		Instance = this;
		PlayMusic();
		_tokens.Add(EventBus.Subscribe<LocalPlayerSpawnedEvent>(OnLocalPlayerSpawned));
		_tokens.Add(EventBus.Subscribe<CameraDroppedEvent>(OnCameraDropped));
		// Отписка от событий при выходе в главное меню
		TreeExiting += () =>
		{
			_tokens.ForEach(token => token.Unsubscribe());
			_tokens.Clear();
		};
		
		// Всё что дальше должно происходить только на сервере или в одиночной игре
		if(!(Network.IsServer || Network.Mode == NetworkMode.None)) return;

		_spawnPool.Add(Game.Storage.Entities.Enemy, 1);
		_spawnPool.Add(Game.Storage.Entities.Flyer, 0.1);
		
		// Во-первых, если это не выделенный сервер, то спавним игрока
		if (Network.IsClient || Network.Mode == NetworkMode.None)
		{
			_localPlayer = SpawnPlayerFor(1);
		}
		
		// Подписываемся на готовность клиента
		_tokens.Add(EventBus.Subscribe<PeerReadyPacket>(OnPeerReady));
		
		// Удаляем игрока при разрыве соединения
		Network.Api.PeerDisconnected += id => GetNode<Character>($"{id}").CleanUp();
	}

	/// <inheritdoc />
	public override void _Process(double delta)
	{
		_gameTime += delta;
		if (Enemies.Count < MaxEnemies)
		{
			SpawnEnemy();
		}
	}

	private void PlayMusic()
	{
		Audio2D.PlayMusic(Music.WorldBgm, 0.5f).Finished += PlayMusic;
	}
	
	

	private Enemy SpawnEnemy()
	{
		var enemy = SelectEnemyType();
		var circumcircle = new Circle(LocalPlayer.Position, SpawnRadius);
		enemy.Position = Rand.OnCircle(circumcircle);
		enemy.Rotation = Rand.Range(Mathf.Tau);

		enemy.Data.MaxHealth *= EnemyHpFactor;
		enemy.Health *= EnemyHpFactor;
		
		enemy.Data.MaxHealth += Instance.EnemyHpOffsetPerDifficulty * Difficulty;
		enemy.Health += Instance.EnemyHpOffsetPerDifficulty * Difficulty;

		enemy.Data.MovementSpeed *= EnemySpeedFactor;
		enemy.Data.MovementSpeed += Instance.EnemySpeedOffsetPerDifficulty * Difficulty;
		
		enemy.ContactDamage *= EnemyDamageFactor;
		enemy.ContactDamage += Instance.EnemyDamageOffsetPerDifficulty * Difficulty;
		
		AddCharacter(enemy);
		return enemy;
	}

	public Enemy SelectEnemyType()
	{
		return _spawnPool.PickRandom().Instantiate() as Enemy;
	}

	// Спавн игроков происходит через этот метод, вызываемый на хосте
	private Player SpawnPlayerFor(NetworkPeer id)
	{
		var player = Game.Storage.Entities.Player.Instantiate() as Player;
		
		// Важно: имя ноды будет синхронизировано даже до _EnterTree. Можем использовать это, чтобы передать управление клиенту.
		//player.ApplyPredefinedComponents();
		player.Name = $"{id}";
		//player.PlayerId = id;
		
		// Добавляем ноду в мир.
		// Обрати внимание, что в этот момент у клиентов тоже создастся такая нода.
		// Это сработает, только если Auto Spawn List в CharacterSpawner содержит путь к этой сцене.
		// CharacterSpawner лежит в Scenes/States/Gameplay/Gameplay.tscn.
		// Он будет синхронизировать наличие известных ему подсцен в целевой ноде (в данном случае в этой).
		AddCharacter(player);
		Log.Info("Built player for peer " + id);
		
		return player;
	}
	
	// Прикручиваем пол и камеру к локальному игроку
	private void OnLocalPlayerSpawned(LocalPlayerSpawnedEvent evt)
	{
		_localPlayer = evt.Player;
		LocalPlayer.SelectWeapon();
	}

	private void OnCameraDropped(CameraDroppedEvent evt)
	{
		var camera = evt.Camera;
		var player = evt.Player;

		if (player.IsLocal)
		{
			Floor.Camera = camera;
			camera.Enabled = true;
			camera.MakeCurrent();
		}
		else
		{
			camera.QueueFree();
		}
	}

	// Если пир смог ссинхронизировать регистри, то он отправляет этот пакет. Если пакет не пришел - новый игрок не спавнится.
	private void OnPeerReady(PeerReadyPacket packet)
	{
		Log.Info($"Player {packet.Sender} ready");
		SpawnPlayerFor(packet.Sender);
	}

	/// <inheritdoc />
	public override void _Input(InputEvent @event)
	{
		if (@event.IsActionPressed(Keys.WheelUp))
		{
			LocalPlayer.Camera.Zoom *= 1.1;
		}

		if (@event.IsActionPressed(Keys.WheelDown))
		{
			LocalPlayer.Camera.Zoom /= 1.1;
		}
	}
}
