﻿using System.Linq;
using Content;
using KludgeBox.Collections;
using VectorSurvivors.Events;

namespace VectorSurvivors;

public partial class World
{
    public static World Instance { get; private set; }

    public static double GameTime
    {
        get => Instance._gameTime;
        set => Instance._gameTime = value;
    }

    public static ReadOnlyHashSet<Character> Characters => Instance._characters.AsReadOnly();
    public static ReadOnlyHashSet<Enemy> Enemies => Instance._enemies.AsReadOnly();
    public static ReadOnlyHashSet<Player> Players => Instance._players.AsReadOnly();
    public static ReadOnlyHashSet<Orb> Orbs => Instance._orbs.AsReadOnly();
    public static Player LocalPlayer => Instance._localPlayer;

    public static double Difficulty => (Instance.BaseDifficulty + GameTime / 60 * Instance.DifficultyPerMinute) * Instance.DifficultyFactor;
    public static int MaxEnemies => Mathf.CeilToInt(Instance.BaseEnemyCount + Instance.EnemyPerDifficulty * Difficulty);
    
    

    public static void AddEntity(Node2D node)
    {
        Instance._entities.Add(node);
        
        if(node is Orb orb)
        {
            Instance._orbs.Add(orb);
            Instance.AddChild(orb, true);
            return;
        }
        
        if(node is Character character)
        {
            AddCharacter(character);
            return;
        }
        
        node.TreeExited += () => Notify_EntityDespawned(node);
        Instance.AddChild(node, true);
    }
    
    private static void AddCharacter(Character character)
    {
        if (!character.IsValid()) return;
        if(EventBus.TryPublish(new CharacterAboutToSpawnEvent(character))) return;
        
        Instance._characters.Add(character);
        Instance.AddChild(character, true);
        
        if(character is Player player)
        {
            Instance._players.Add(player);
        }

        if(character is Enemy enemy)
        {
            Instance._enemies.Add(enemy);
        }
        
        EventBus.Publish(new CharacterSpawnedEvent(character));
    }
    
    
    public static void RemoveCharacter(Character character)
    {
        if(!character.IsValid()) return;
        if(EventBus.TryPublish(new CharacterAboutToDespawnEvent(character))) return;
        
        Instance._characters.Remove(character);
        
        if (character is Player player)
        {
            Instance._players.Remove(player);
        }
        if (character is Enemy enemy)
        {
            Instance._enemies.Remove(enemy);
            
        }
        character.QueueFree();
        
        EventBus.Publish(new CharacterDespawnedEvent(character));
    }
    
    public static void ExitWorld()
    {
        Instance.QueueFree();
        Audio2D.StopMusic(1);
        Game.State.ChangeTo(Game.Storage.States.MainMenuState);
    }
    
    internal static void Notify_CharacterDied(Character character)
    {
        if (character is Enemy enemy)
        {
            Audio2D.PlaySoundAt(Sfx.Pop, enemy.Position).PitchVariation(0.2f);
            
            if(!Players.Any(player => player.CanLevelUp)) return;
            
            var orb = XpOrb.Create(enemy.XpReward);
            orb.Position = enemy.Position;
            orb.Velocity = orb.MaxSpeed * 0.2;
            orb.Direction = Rand.UnitVector;
            
            AddEntity(orb);
        }

        if (character == LocalPlayer)
        {
            ExitWorld();
        }
        RemoveCharacter(character);
    }
    
    internal static void Notify_EntityDespawned(Node2D entity)
    {
        Instance._entities.Remove(entity);
        
        if(entity is Character character)
        {
            RemoveCharacter(character);
        }
        
        if(entity is Orb orb)
        {
            Instance._orbs.Remove(orb);
        }
    }
}