﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;

namespace VectorSurvivors;

[AttributeUsage(AttributeTargets.Property)]
public class WorldProperty : Attribute
{
    public string Name { get; private set; }
    public WorldProperty(){}
    public WorldProperty(string name) => Name = name;
}

public partial class World
{
    [ExportGroup("World Settings")]
    [ExportSubgroup("Movement")]
    
    [WorldProperty("Трение")]
    [Export] public double Friction { get; set; } = 1;
    
    
    [ExportSubgroup("Progression")]
    // Progression
    // Xp required for level up calculated as
    // BaseRequiredXp * (RequiredXpLevelFactor ^ Level) + (RequiredXpLevelOffset * Level)
    
    [WorldProperty("Базовое значение требуемого опыта")]
    [Export] public int BaseRequiredXp { get; set; } = 5; // How much XP is required to level up from level 1 to level 2
    
    [WorldProperty("Дополнительный требуемый опыт за уровень")]
    [Export] public double RequiredXpLevelOffset { get; set; } = 1; // How much more XP will be required at each level
    
    [WorldProperty("Множитель дополнительного требуемого опыта за уровень")]
    [Export] public double RequiredXpLevelFactor { get; set; } = 1.0; // How much more XP will be required at each level
    
    [WorldProperty("Множитель получаемого опыта")]
    [Export] public double XpYieldFactor { get; set; } = 1.0; // How much XP is yielded when killing an enemy
    
    // When player getting xp, every other ally player will get this much of it. The more players there are, the more XP will get the team.
    [Export] public double XpShareFactor { get; set; } = 0.5; 
    
    [WorldProperty("Макс. уровень игрока")]
    [Export] public int MaxPlayerLevel { get; set; } = 150;
    
    [WorldProperty("Макс. уровень оружия")]
    [Export] public int MaxWeaponLevel { get; set; } = 50;
    
    [WorldProperty("Макс. слотов оружия")]
    [Export] public int MaxWeaponSlots { get; set; } = 4;
    
    [WorldProperty("Улучшения за уровень")]
    [Export] public int UpgradesPerLevel { get; set; } = 3;
    
    [WorldProperty("Макс. рероллов за уровень")]
    [Export] public int MaxRerollsOnLevelUp { get; set; } = 2;
    
    [WorldProperty("Рероллов за сброс")]
    [Export] public int RerollsPerDrop { get; set; } = 1;
    
    
    [ExportSubgroup("Difficulty")]
    
    [WorldProperty("Начальная сложность")]
    [Export] public double BaseDifficulty { get; set; } = 0.0;
    
    [WorldProperty("Множитель сложности")]
    [Export] public double DifficultyFactor { get; set; } = 1.0;
    
    [WorldProperty("Рост сложности в минуту")]
    [Export] public double DifficultyPerMinute { get; set; } = 1.0;
    
    
    [WorldProperty("Начальное количество врагов")]
    [Export] public int BaseEnemyCount { get; set; } = 20;
    
    [WorldProperty("Дополнительно врагов за единицу сложности")]
    [Export] public int EnemyPerDifficulty { get; set; } = 7;
    
    [WorldProperty("Радиус появления врагов")]
    [Export] public double SpawnRadius { get; set; } = 1300.0;
    
    
    [WorldProperty("Множитель здоровья врагов")]
    [Export] public double EnemyHpFactor { get; set; } = 1;
    
    [WorldProperty("Дополнительное здоровье врагов за единицу сложности")]
    [Export] public double EnemyHpOffsetPerDifficulty { get; set; } = 20.0;
    
    
    [WorldProperty("Множитель скорости врагов")]
    [Export] public double EnemySpeedFactor { get; set; } = 1;
    
    [WorldProperty("Дополнительная скорость врагов за единицу сложности")]
    [Export] public double EnemySpeedOffsetPerDifficulty { get; set; } = 10;
    
    [WorldProperty("Множитель урона врагов")]
    [Export] public double EnemyDamageFactor { get; set; } = 1;
    
    [WorldProperty("Дополнительный урон врагов за единицу сложности")]
    [Export] public double EnemyDamageOffsetPerDifficulty { get; set; } = 1;
    
    public List<WorldPropertyHandle> GetWorldProperties()
    {
        var properties = GetType()
            .GetProperties()
            .Where(x => x.GetCustomAttribute<WorldProperty>() != null);
        return properties.Select(x => new WorldPropertyHandle(x, this)).ToList();
    }
}

public record WorldPropertyHandle
{
    public object Value
    {
        get => Property.GetValue(World);
        set => Property.SetValue(World, ForceRecast(value));
    }

    public PropertyInfo Property { get; init; }
    public World World { get; init; }
    public WorldPropertyHandle(PropertyInfo property, World world)
    {
        Property = property;
        World = world;
    }

    public string Name => Property.GetCustomAttribute<WorldProperty>()!.Name;
    public Type Type => Property.PropertyType;
    public PropertyControl AsControl()
    {
        var property = Game.Storage.Ui.WorldProperty.Instantiate() as PropertyControl;
        property.Initialize(this);
        return property;
    }
    
    

    private object ForceRecast(object value)
    {
        if (Type == typeof(int))
        {
            return Convert.ToInt32(value);
        }
        if (Type == typeof(double))
        {
            return Convert.ToDouble(value);
        }
        return value;
    }

    public void Deconstruct(out PropertyInfo Property, out World World)
    {
        Property = this.Property;
        World = this.World;
    }
}