using System;
using System.Collections.Generic;
using System.Linq;
using Content;

namespace VectorSurvivors;

public partial class ProjectileWeapon : Weapon
{
	[Export] public double Range { get; set; } = 500;
	[Export] public double ProjectileSpeed { get; set; } = 1000;
	[Export] public double ProjectileLifetime { get; set; } = 2;
	[Export] public bool IsPiercing { get; set; } = false;
	
	
	[Export] public double CritChance { get; set; } = 0.01;
	[Export] public double CritMultiplier { get; set; } = 1.5;
	
	[Export] public int ProjectilesPerShot { get; set; } = 1;
	[Export] public int ShootCount { get; set; } = 2;
	[Export] public double ShootSpread { get; set; } = Mathf.Pi / 6;
	[Export] public int ProjectilesPerAttackThreshold { get; set; } = 9;
	
	public int ProjectilesPerAttack => ProjectilesPerShot * ShootCount;
	public double ShotDelay => 0.15 * ReloadTime;
	public override Texture2D Icon => Game.Storage.Textures.ProjectileWeaponIcon.Duplicate() as Texture2D;
	public override string StateInfo => $"Урон: {BaseDamage:N0}\n" +
	                                    $"Крит. урон: +{CritMultiplier * 100:N0}%\n" +
	                                    $"Крит. шанс: {CritChance * 100:N0}%\n" +
	                                    //$"Range: {Range:N0}\n" +
	                                    $"Перезарядка: {ReloadTime:N2} сек.\n" +
	                                    $"Пробитие: {(IsPiercing ? "да" : "нет")}\n" +
	                                    $"Снарядов: {ProjectilesPerShot}\n" +
	                                    $"Выстрелов: {ShootCount}\n" +
	                                    $"Скорость снаряда: {ProjectileSpeed:N0}";
	
	private double _cooldown = 0;
	private double _shotCooldown = 0;
	private double _shotsLeft = 1;
	

	public ProjectileWeapon()
	{
		
	}
	
	/// <inheritdoc />
	public override void _Ready()
	{
		//UpgradesRepository.AddToAllExcept(UpgradeRange, Tier.Legendary, Tier.Epic);
		
		UpgradesRepository.AddToAll(UpgradeDamage);
		UpgradesRepository.AddToAll(UpgradeSpeed);
		UpgradesRepository.AddToAll(UpgradeReloadTime);
		UpgradesRepository.AddToAll(UpgradeCritChance);
		UpgradesRepository.AddToAll(UpgradeCritMultiplier);
		
		UpgradesRepository.Add(UpgradePiercing, Tier.Legendary);
		UpgradesRepository.Add(UpgradeProjectilesCount, Tier.Legendary);
		UpgradesRepository.Add(UpgradeShotsCount, Tier.Legendary);
		
		base._Ready();
	}

	


	public override PendingUpgrade GetUpgrade(Player player, Tier tier)
	{
		var result = base.GetUpgrade(player, tier);
		_getUpgradeAttempts = 0;

		return result;
	}
	
	/// <inheritdoc />
	public override void _Process(double delta)
	{
		base._Process(delta);
	}

	public virtual PiercingProjectile CreateProjectile()
	{
		return PiercingProjectile.Create();
	}

	public virtual void PlayShootSound()
	{
		Audio2D.PlaySoundAt(Sfx.SmallLaserShot, GlobalPosition);
	}

	/// <inheritdoc />
	public override void UpdateAttack(Player player, double delta)
	{
		if (_shotsLeft > 0)
		{
			_cooldown = 0;
			_shotCooldown += delta;
			if (_shotCooldown >= ShotDelay)
			{
				_shotsLeft--;
				_shotCooldown -= ShotDelay;
				Shoot(player, ProjectilesPerShot);
			}
		}
		else
		{
			_cooldown += delta;
			if (_cooldown >= ReloadTime)
			{
				_cooldown -= ReloadTime;
				_shotsLeft = ShootCount;
				_shotCooldown = 0;
			}
		}
	}

	private void Shoot(Player player, int amount = 1)
	{
		amount = Math.Max(amount, 1);
		Enemy target = World.Enemies.MinBy(enemy => (enemy.Position - player.Position).LengthSquared());

		var totalRange = Range * player.UniversalAttackDamageFactor;
		if((target.Position - player.Position).LengthSquared() > totalRange * totalRange) return;
		
		for (int i = 0; i < amount; i++)
		{
			var spread = ShootSpread;
			var anglePerStep = spread / (amount - 1);
			var angle = amount == 1 ? 0 : -spread * 0.5 + anglePerStep * i;
			
			
			var projectile = CreateProjectile();
			var rotation = (target.Position - player.Position).Angle() + angle;
		
			projectile.Position = player.Position;
			projectile.Rotation = rotation;

			ConfigureProjectile(player, projectile);
		
			projectile.Configure(this, player);
		
			World.Instance.AddChild(projectile);
		}
		PlayShootSound();
	}

	public virtual void ConfigureProjectile(Player player, PiercingProjectile projectile)
	{
		projectile.Damage = BaseDamage * player.UniversalAttackDamageFactor;
		projectile.Lifetime = ProjectileLifetime;
		projectile.Speed = ProjectileSpeed;
		projectile.CritChance = CritChance;
		projectile.CritMultiplier = CritMultiplier;
		projectile.IsPiercing = IsPiercing;
	}


	protected virtual double GetDamageUpgradeMultiplier(Player player, Tier tier)
	{
		return UpgradeUtils.SwitchTier(tier,
			common: 0.05,
			uncommon: 0.1,
			rare: 0.2,
			epic: 0.3,
			legendary: 0.5
		);
	}
	private PendingUpgrade UpgradeDamage(Player player, Tier tier)
	{
		var tieredFactor = GetDamageUpgradeMultiplier(player, tier);
		var totalMultiplier = 1 + tieredFactor;
		var icon = Icon;

		var percentage = tieredFactor * 100;
		var newDamage = Mathf.Ceil(totalMultiplier * BaseDamage);
		
		var name = $"Урон +{percentage:N0}%";
		var description = $"Увеличивает урон снарядов\n{BaseDamage:N0} -> {newDamage:N0}";
		
		return new PendingUpgrade(name, description, tier, this, ()=> BaseDamage = newDamage, icon);
	}
	

	protected virtual double GetCritChanceUpgradeOffset(Player player, Tier tier)
	{
		return (int)tier switch
		{
			1 => 2,
			2 => 4,
			3 => 6,
			4 => 10,
			5 => 15
		};
	}
	private PendingUpgrade UpgradeCritChance(Player player, Tier tier)
	{
		var baseValue = 0.01;
		var tierMultiplier = GetCritChanceUpgradeOffset(player, tier);
		var tieredValue = baseValue * tierMultiplier;
		var icon = Icon;
		var percentage = tieredValue * 100;
		var newChance = CritChance + tieredValue;
		
		var name = $"Крит. шанс +{percentage}%";
		var description = $"Увеличивает шанс критического урона\n{CritChance*100:N2}% -> {newChance*100:N2}%";
		
		return new PendingUpgrade(name, description, tier, this, ()=> CritChance = newChance, icon);
	}
	
	protected virtual double GetCritMultiplierUpgradeOffset(Player player, Tier tier)
	{
		return (int)tier switch
		{
			1 => 5,
			2 => 10,
			3 => 20,
			4 => 35,
			5 => 50
		};
	}
	
	private PendingUpgrade UpgradeCritMultiplier(Player player, Tier tier)
	{
		var baseValue = 0.01;
		var tierMultiplier = GetCritMultiplierUpgradeOffset(player, tier);
		var tieredValue = baseValue * tierMultiplier;
		var icon = Icon;
		var percentage = tieredValue * 100;
		var newMultiplier = CritMultiplier + tieredValue;
		
		var name = $"Крит. урон +{percentage}%";
		var description = $"Увеличивает множитель критического урона\n{CritMultiplier*100:N0}% -> {newMultiplier*100:N0}%";
		
		return new PendingUpgrade(name, description, tier, this, ()=> CritMultiplier = newMultiplier, icon);
	}
	
	private PendingUpgrade UpgradeSpeed(Player player, Tier tier)
	{
		var baseFactor = 0.05;
		var tieredFactor = baseFactor * tier;
		var totalMultiplier = 1 + tieredFactor;
		var icon = Icon;
		
		var percentage = tieredFactor * 100;
		var newSpeed = ProjectileSpeed * totalMultiplier;
		
		var name = $"Скорость снаряда +{percentage:N0}%";
		var description = $"Увеличивает скорость полета снаряда\n{ProjectileSpeed:N0} -> {newSpeed:N0}";
		
		return new PendingUpgrade(name, description, tier, this, ()=> ProjectileSpeed = newSpeed, icon);
	}

	protected virtual double GetReloadTimeUpgradeMultiplier(Player player, Tier tier)
	{
		return UpgradeUtils.SwitchTier(tier,
			common: 0.05,
			uncommon: 0.1,
			rare: 0.15,
			epic: 0.25,
			legendary: 0.5);
	}
	private PendingUpgrade UpgradeReloadTime(Player player, Tier tier)
	{
		var baseFactor = 0.05;
		var tieredFactor = GetReloadTimeUpgradeMultiplier(player, tier);
		var totalMultiplier = 1 - tieredFactor;
		var icon = Icon;
		
		var percentage = tieredFactor * 100;
		var newReloadTime = ReloadTime * totalMultiplier;
		
		var name = $"Перезарядка -{percentage:N0}%";
		var description = $"Уменьшает время перезарядки снаряда\n{ReloadTime:N2} -> {newReloadTime:N2} cек";
		
		return new PendingUpgrade(name, description, tier, this, ()=> ReloadTime = newReloadTime, icon);
	}
	
	private PendingUpgrade UpgradeProjectilesCount(Player player, Tier tier)
	{
		if (ProjectilesPerAttack >= ProjectilesPerAttackThreshold) return TryGetAnotherUpgrade(player, tier);
		
		var tierMult = Mathf.Min(tier - 3, 1);
		var addProjectiles = 2 * tierMult;
		var icon = Icon;
		
		tier = Tier.Unique;
		var newProjectilesCount = ProjectilesPerShot + addProjectiles;
		var name = $"Снаряды +{addProjectiles:N0}";
		var description = $"Выпускает\n{ProjectilesPerShot} -> {newProjectilesCount:N0} {TextUtils.Declension(newProjectilesCount, "снаряд", "снаряда", "снарядов")} каждый выстрел";
		
		return new PendingUpgrade(name, description, tier, this, ()=> ProjectilesPerShot += addProjectiles, icon);
	}
	
	private PendingUpgrade UpgradeShotsCount(Player player, Tier tier)
	{
		if (ProjectilesPerAttack >= ProjectilesPerAttackThreshold) return TryGetAnotherUpgrade(player, tier);
		var shots = 1;
		
		tier = Tier.Unique;
		var newShotsCount = ShootCount + shots;
		var name = $"Выстрелы +{shots:N0}";
		var description = $"Делает\n{ShootCount:N0} -> {newShotsCount:N0} {TextUtils.Declension(newShotsCount, "выстрел", "выстрела", "выстрелов")}";
		var icon = Icon;
		
		return new PendingUpgrade(name, description, tier, this, ()=> ShootCount += shots, icon);
	}
	
	private PendingUpgrade UpgradePiercing(Player player, Tier tier)
	{
		if (IsPiercing)
			return TryGetAnotherUpgrade(player, tier);
		
		var icon = Icon;
		
		var name = $"Пробитие";
		var description = $"Снаряд не уничтожается при попадании и продолжает лететь сквозь врагов до истечения времени существования";
		tier = Tier.Unique;
		
		return new PendingUpgrade(name, description, tier, this, () =>
		{
			IsPiercing = true;
		}, icon);
	}

	private int _getUpgradeAttempts = 0;
	private PendingUpgrade TryGetAnotherUpgrade(Player player, Tier tier)
	{
		if (_getUpgradeAttempts > 5)
		{
			return null;
		}
		
		_getUpgradeAttempts++;
		return GetUpgrade(player, tier);
	}
	
}