﻿using System;
using System.Collections.Generic;
using System.Linq;
using KludgeBox.Collections;

namespace VectorSurvivors;

public partial class WeaponsInventory : Node2D
{
    [Export] [NotNull] protected Player Player { get; set; }
    
    public ReadOnlyHashSet<Weapon> Weapons => _weapons.AsReadOnly();
    private HashSet<Weapon> _weapons = new();
    

    /// <inheritdoc />
    public override void _Ready()
    {
        NotNullChecker.CheckProperties(this);
        var children = GetChildren();
        
        foreach (Node child in children)
        {
            if(child is Weapon weapon)
            {
                RegisterWeapon(weapon);
            }
            else
            {
                Log.Warning($"{child} is not a weapon");
            }
        }

        ChildEnteredTree += node =>
        {
            if (_weapons.Contains(node)) return;
            
            RemoveChild(node);
            Log.Warning($"Removed unauthorized node '{node}' from weapons inventory");
        };

        ChildExitingTree += node =>
        {
            if (_weapons.Contains(node))
            {
                _weapons.Remove(node as Weapon);
                Log.Warning($"Weapon '{node}' was removed from inventory unexpectedly");
            }
        };
    }

    public void Update(Player player, double delta)
    {
        foreach (Weapon weapon in _weapons)
        {
            weapon.UpdateAttack(player, delta);
        }
    }

    public bool HasWeapon<TWeapon>() where TWeapon : Weapon
    {
        var type = typeof(TWeapon);

        return _weapons.Any(weapon => weapon.GetType() == type);
    }
    
    public bool HasWeapon(Type type)
    {
        return _weapons.Any(weapon => weapon.GetType() == type);
    }

    public TWeapon GetWeapon<TWeapon>() where TWeapon : Weapon
    {
        var type = typeof(TWeapon);
        return GetWeapon(type) as TWeapon;
    }
    
    public Weapon GetWeapon(Type type)
    {
        return _weapons.FirstOrDefault(weapon => weapon.GetType() == type);
    }
    
    public bool CanAddWeapon(Weapon weapon)
    {
        return (_weapons.Count < World.Instance.MaxWeaponSlots) && (!HasWeapon(weapon.GetType()));
    }
    
    public bool CanAddWeapon()
    {
        return _weapons.Count < World.Instance.MaxWeaponSlots && GetMissingWeaponTypes().Count > 0;
    }

    public void RemoveWeapon(Weapon weapon)
    {
        if(!_weapons.Contains(weapon))
            return;

        weapon.OnUnequipped(Player);
        _weapons.Remove(weapon);
        RemoveChild(weapon);
    }
    
    public void AddWeapon(Weapon weapon)
    {
        if(RegisterWeapon(weapon))
            AddChild(weapon);
    }
    
    public List<Type> GetMissingWeaponTypes()
    {
        return Game.Storage.Weapons.ByType.Keys.Except(GetContainedWeaponTypes()).ToList();
    }
    
    public List<Type> GetContainedWeaponTypes()
    {
        return _weapons.Select(weapon => weapon.GetType()).ToList();
    }
    
    private bool RegisterWeapon(Weapon weapon)
    {
        bool registered = _weapons.Add(weapon);
        
        if(registered)
            weapon.OnEquipped(Player);
            
        return registered;
    }
}