using Godot;
using System;
using VectorSurvivors;

public partial class GravityFieldComponent : CharacterComponent
{
	public Wormhole Source { get; private set; }
	public GpuParticles2D Particles { get; private set; }

	public double Lifetime { get; set; } = 1;
	
	public GravityFieldComponent(Wormhole source)
	{
		Source = source;
	}
	
	
	/// <inheritdoc />
	public override void OnProcess(Character character, double delta)
	{
		Lifetime -= delta;
		var rangeSquared = Source.FieldRadius * Source.FieldRadius;
		
		if (Lifetime > 0 
		    && Source.IsValid() && !Source.IsQueuedForDeletion()
		    && character.IsValid() && !character.IsQueuedForDeletion()
		    && character.DistanceSquaredTo(Source) <= rangeSquared)
		{
			var dir = character.DirectionTo(Source);
			character.Position += dir * Source.GravityStrength * delta;
			Particles.Position = character.Position;
			
			var material = Particles.ProcessMaterial as ParticleProcessMaterial;
			var dist = character.DistanceTo(Source);
			var accel = (2 * dist) / (Particles.Lifetime * Particles.Lifetime);
			var gravity = dir * accel;
			material.Gravity = new Vector3(gravity.X, gravity.Y, 0);
		}
		else
		{
			character.RemoveComponent(this);
		}
	}

	/// <inheritdoc />
	public override void OnAdded(Character character)
	{
		Particles = new GpuParticles2D();
		var material = new ParticleProcessMaterial();
		
		Particles.ProcessMaterial = material;
		Particles.TrailEnabled = true;
		Particles.TrailLifetime = 0.1;
		Particles.Lifetime = 1;
		Particles.Amount = 24;
		Particles.Position = character.Position;
		Particles.ZIndex = 2;

		var colorRamp = new GradientTexture1D();
		colorRamp.Gradient = new Gradient();
		colorRamp.Gradient.Colors =
		[
			Col(0), 
			Col(0.5f, 0, 1f, 1),
			Col(0.5f, 0, 1f, 0)
		];
		colorRamp.Gradient.Offsets =
		[
			0,
			0.5f,
			1
		];
		material.ColorRamp = colorRamp;
		material.EmissionShape = ParticleProcessMaterial.EmissionShapeEnum.Sphere;
		material.EmissionSphereRadius = 35f;
		material.ScaleMin = material.ScaleMax = 3;
		
		World.AddEntity(Particles);
	}

	/// <inheritdoc />
	public override void OnRemove(Character character)
	{
		Particles.Emitting = false;
		Particles.Destruct(Particles.Lifetime);
	}

	/// <inheritdoc />
	public override void AdjustUniversalAttackDamageFactor(Character character, ref double factor)
	{
		factor = 0;
	}

	/// <inheritdoc />
	public override void AdjustMovementSpeed(Character character, ref double speed)
	{
		speed = 0;
	}

	/// <inheritdoc />
	public override void AdjustRotationSpeed(Character character, ref double rotationSpeed)
	{
		rotationSpeed = 0;
	}

	/// <inheritdoc />
	public override void TakingDamage(Character target, Damage damage)
	{
		damage.Amount *= Source.IncomingDamageMult;
	}
}
