using System.Collections.Generic;
using System.Linq;
using KludgeBox.Scheduling;
using VectorSurvivors;

public partial class Wormhole : Node2D
{
	[Export] [NotNull] public Sprite2D Sprite { get; private set; }
	[Export] [NotNull] public Sprite2D DistortionSprite { get; private set; }
	[Export] [NotNull] public CpuParticles2D VortexParticles { get; private set; }
	
	
	[Export] public double FieldRadius { get; set; } = 200;
	[Export] public double Lifetime { get; set; } = 5;
	[Export] public double BaseDamage { get; set; } = 5;
	[Export] public double DamageInterval { get; set; } = 1.5;
	[Export] public double MinDamageMultiplier { get; set; } = 0.5;
	[Export] public double MaxDamageMultiplier { get; set; } = 1;
	[Export] public double IncomingDamageMult { get; set; } = 1;
	[Export] public double GravityStrength { get; set; } = 1;

	public Player Inflicitor { get; set; }
	
	private double _timeToDamage = 0;
	private double _timeToLive = 0;
	private Cooldown _tickCooldown = new Cooldown(0.05);
	private Enemy[] _affected;
	

	public Wormhole()
	{
		_tickCooldown.Ready += OnTick;
	}

	/// <inheritdoc />
	public override void _Ready()
	{
		DistortionSprite.Modulate = Col(1, 1, 1, 0);
		var tween = DistortionSprite.CreateTween();
		tween.TweenProperty(DistortionSprite, "modulate:a", 1, 1);
	}

	/// <inheritdoc />
	public override void _Process(double delta)
	{
		_tickCooldown.Update(delta);
		_timeToDamage += delta;
		_timeToLive += delta;
		
		if (_timeToDamage >= DamageInterval)
		{
			_timeToDamage -= DamageInterval;
			foreach (Enemy enemy in _affected)
			{
				var distance = enemy.DistanceTo(this);
				var damageMult = Mathf.Remap(distance, 0, FieldRadius, MinDamageMultiplier, MaxDamageMultiplier);
				var damage = new Damage(BaseDamage * damageMult* Inflicitor.UniversalAttackDamageFactor, false, false, Inflicitor);
				enemy.TakeDamage(damage);
			}
		}
		
		if (_timeToLive >= Lifetime)
		{
			DistortionSprite.Destruct(1);
			var tween = DistortionSprite.CreateTween();
			tween.TweenProperty(DistortionSprite, "modulate:a", 0, 1);
			DistortionSprite.Drop();

			//VortexParticles.Drop(true);
			VortexParticles.Emitting = false;
			VortexParticles.Destruct(VortexParticles.Lifetime);
			
			QueueFree();
		}
	}
	
	private void OnTick()
	{
		var rangeSquared = FieldRadius * FieldRadius;
		_affected = World.Enemies.Where(enemy => enemy.DistanceSquaredTo(this) <= rangeSquared).ToArray();
		foreach (Enemy enemy in _affected)
		{
			if(!enemy.IsValid() || enemy.IsQueuedForDeletion()) continue;
			RefreshGravity(enemy);
		}
	}

	private void RefreshGravity(Enemy enemy)
	{
		if (enemy.TryGetComponent<GravityFieldComponent>(out var component))
		{
			component.Lifetime = 1;
		}
		else
		{
			var comp = new GravityFieldComponent(this);
			enemy.AddComponent(comp);
		}
	}
}
