using Godot;
using System;
using Content;
using VectorSurvivors;

public partial class WormholeWeapon : Weapon
{
	[Export] public double FieldRadius { get; set; } = 200;
	[Export] public double Lifetime { get; set; } = 5;
	[Export] public double DamageInterval { get; set; } = 1.5;
	[Export] public double MinDamageMultiplier { get; set; } = 0.5;
	[Export] public double MaxDamageMultiplier { get; set; } = 1;
	[Export] public double GravityStrength { get; set; } = 40;
	[Export] public double IncomingDamageMult { get; set; } = 1;
	[Export] public double SpawnRadius { get; set; } = 700;

	/// <inheritdoc />
	public override Texture2D Icon => Game.Storage.Textures.WormholeWeaponIcon;

	public override string StateInfo =>
		$"Базовый урон: {BaseDamage:N0}\n" +
		$"Урон: {BaseDamage * MinDamageMultiplier:N0} - {BaseDamage * MaxDamageMultiplier:N0}\n" +
		$"Интервал урона: {DamageInterval:N2} сек.\n" +
		$"Перезарядка: {ReloadTime:N0} сек.\n" +
		$"Радиус: {FieldRadius:N0}\n" +
		$"Ослабление: {IncomingDamageMult:N2}\n" +
		$"Время жизни: {Lifetime:N1}";

	private double _spawnCooldown = 0;
	private Wormhole _wormhole;
	
	public override void _Ready()
	{
		//UpgradesRepository.AddToAllExcept(UpgradeRange, Tier.Legendary, Tier.Epic);
		
		UpgradesRepository.AddToAll(UpgradeDamage);
		UpgradesRepository.AddToAll(UpgradeDamageInterval);
		UpgradesRepository.AddToAll(UpgradeReloadTime);
		UpgradesRepository.AddToAll(UpgradeRadius);
		UpgradesRepository.AddToAll(UpgradeLifetime);
		
		UpgradesRepository.Add(UpgradeWeakening, Tier.Legendary, Tier.Epic);
		
		base._Ready();
	}
	
	
	public override PendingUpgrade GetUpgrade(Player player, Tier tier)
	{
		var result = base.GetUpgrade(player, tier);
		_getUpgradeAttempts = 0;

		return result;
	}
	/// <inheritdoc />
	public override void UpdateAttack(Player player, double delta)
	{
		if(_wormhole.IsValid()) return;
		
		_spawnCooldown += delta;
		if (_spawnCooldown >= ReloadTime)
		{
			_spawnCooldown -= ReloadTime;
			SpawnWormhole(player);
		}
	}

	private void SpawnWormhole(Player player)
	{
		var wormhole = Game.Storage.Entities.Wormhole.Instantiate() as Wormhole;
		wormhole.Position = Rand.InsideCircle(new(player.Position, SpawnRadius));
		wormhole.Lifetime = Lifetime;
		wormhole.FieldRadius = FieldRadius;
		wormhole.IncomingDamageMult = IncomingDamageMult;
		wormhole.MinDamageMultiplier = MinDamageMultiplier;
		wormhole.MaxDamageMultiplier = MaxDamageMultiplier;
		wormhole.GravityStrength = GravityStrength;
		wormhole.DamageInterval = DamageInterval;
		wormhole.BaseDamage = BaseDamage;
		wormhole.Inflicitor = player;
		wormhole.Scale = Vec(FieldRadius / 200);
		
		World.AddEntity(wormhole);
		_wormhole = wormhole;
		Audio2D.PlaySoundAt(Sfx.ElectricHorn, wormhole.Position, 2).PitchVariation(0.2f);
	}

	private PendingUpgrade UpgradeWeakening(Player player, Tier tier)
	{
		tier = Tier.Unique;
		var offset = 0.5;
		var icon = Icon;
		var newIncomingDamageMult = IncomingDamageMult + offset;

		var name = $"Ослабление +{offset*100:N0}%";
		var description = $"Увеличивает весь урон по врагам, попавшим в поле тяготения червоточины\n{IncomingDamageMult*100:N0}% -> {newIncomingDamageMult*100:N0}%";
		
		return new PendingUpgrade(name, description, tier, this, ()=> IncomingDamageMult = newIncomingDamageMult, icon);
	}

	private PendingUpgrade UpgradeLifetime(Player player, Tier tier)
	{
		var offset = UpgradeUtils.SwitchTier(tier,
			common: 0.1,
			uncommon: 0.2,
			rare: 0.3,
			epic: 0.5,
			legendary: 1);
		
		var icon = Icon;
		
		var newLifetime = Lifetime + offset;
		
		var name = $"Время жизни +{offset:N1}";
		var description = $"Увеличивает время жизни червоточины\n{Lifetime:N1} -> {newLifetime:N1}";
		
		return new PendingUpgrade(name, description, tier, this, ()=> Lifetime = newLifetime, icon);
	}

	private PendingUpgrade UpgradeRadius(Player player, Tier tier)
	{
		var offset = UpgradeUtils.SwitchTier(tier,
			common: 5,
			uncommon: 10,
			rare: 15,
			epic: 25,
			legendary: 50);

		var icon = Icon;
		
		var newRadius = FieldRadius + offset;
		
		var name = $"Радиус +{offset:N0}";
		var description = $"Увеличивает радиус\n{FieldRadius:N0} -> {newRadius:N0}";
		
		return new PendingUpgrade(name, description, tier, this, ()=> FieldRadius = newRadius, icon);
	}

	private PendingUpgrade UpgradeReloadTime(Player player, Tier tier)
	{
		if (ReloadTime <= 5) 
			return TryGetAnotherUpgrade(player, tier);

		int reloadTimeOffset = tier;
		
		var newReloadTime = Mathf.Max(ReloadTime - reloadTimeOffset, 5);
		var icon = Icon;
		var name = $"Перезарядка -{reloadTimeOffset:N0} сек";
		var description = $"Уменьшает время между созданием червоточин {ReloadTime:N0} -> {newReloadTime:N0} сек";
		
		return new PendingUpgrade(name, description, tier, this, ()=> ReloadTime = newReloadTime, icon);
	}

	private PendingUpgrade UpgradeDamageInterval(Player player, Tier tier)
	{
		var multiplier = UpgradeUtils.SwitchTier(tier,
			common: 0.05,
			uncommon: 0.1,
			rare: 0.15,
			epic: 0.2,
			legendary: 0.3);
		
		var totalMultiplier = 1 - multiplier;
		var icon = Icon;
		
		var percentage = multiplier * 100;
		var newInterval = DamageInterval * totalMultiplier;
		
		var name = $"Интервал урона -{percentage:N0}%";
		var description = $"Интервал между нанесением урона урона\n{DamageInterval:N2} -> {newInterval:N2} сек.";
		
		return new PendingUpgrade(name, description, tier, this, ()=> DamageInterval = newInterval, icon);
	}

	private PendingUpgrade UpgradeDamage(Player player, Tier tier)
	{
		var multiplier = UpgradeUtils.SwitchTier(tier,
			common: 0.02,
			uncommon: 0.05,
			rare: 0.07,
			epic: 0.15,
			legendary: 0.25);
		
		var totalMultiplier = 1 + multiplier;
		var icon = Icon;
		
		var percentage = multiplier * 100;
		var newDamage = Mathf.Ceil(BaseDamage * totalMultiplier);
		
		var name = $"Урон +{percentage:N0}%";
		var description = $"Базовый урон\n{BaseDamage:N0} -> {newDamage:N0}";
		
		return new PendingUpgrade(name, description, tier, this, ()=> BaseDamage = newDamage, icon);
	}

	
	private int _getUpgradeAttempts = 0;
	private PendingUpgrade TryGetAnotherUpgrade(Player player, Tier tier)
	{
		if (_getUpgradeAttempts > 5)
		{
			return null;
		}
		
		_getUpgradeAttempts++;
		return GetUpgrade(player, tier);
	}
}
